/*****************************************************
 * 汎用データロガー
 *
 * 多チャンネルのアナログ信号を AD 変換して、カンマ区切りの CSV 文字列
 * 形式で UART 送信します。 
 *
 * 1. AD変換に汎用性があります。
 *      AD 変換サンプリング周波数と、入力チャンネル数、アナログ入力形式
 *      (単相、差動入力)は自由に可変できます。 TopDesign.cysch
 *      の AD変換コンポーネントの設定を変えてください。また、main.h 内
 *      の #define ADC_CHANNEL に入力チャンネル数を記述します。
 *
 * 2. AD変換の開始/停止は、PC からのUART送信で制御されます。
 *      PC からのコマンドは、簡略的にアスキーコード1文字とします。 main.h
 *      内の #define COMMAND_START, #define COMMAND_STOP
 *      に定義されます。AD変換の開始と、停止の2種類のコマンドになります。
 *
 * 3. AD変換が開始されると、PC へ AD変換結果が CSV 文字列で UART 送信されます。
 *      チャンネル数に自由度を持たせるため、AD変換結果はカンマ区切りの CSV
 *      文字列形式で UART 送信します。1回のサンプリングデータを、チャンネル0
 *      を先頭にして、チャンネル数だけ連ね、行末は改行文字('\n') 1文字
 *      で区切ります。
 *
 * 本サンプルプログラムでは、
 *      AD変換 : サンプリング周波数1000zHz, 2ch の単相入力としています。
 *
 * Copyright (C) 2019, CQ Publishing Co.,Ltd., all right reserved.
 *
 * 本ソースコードはフリー・ソフトウェアであり、また無保証です。
 * 用途に制限はありません。あなたの責任の下において、個人的・非営利的な
 * ものから商用製品の開発に及ぶ目的に使用・改変・再配布することができます。
 ****************************************************/
#include <stdio.h>
#include "project.h"
#include "main.h"    /* 自作ヘッダ */

/* AD変換結果の格納変数 */
volatile int16 adc_result[ADC_CHANNEL];
/* AD変換スタンバイ時0, AD変換完了時 > 0 とするフラグ */
volatile int8  adc_done_flag = 0;


int main(void)
{
    int16 i, str_index;
    uint8 command;
    uint8 state = STATE_STOP;             /* 現在のステート。ステートは2種類 STATE_START, STATE_STOP */
    char8 send_message[ADC_CHANNEL*7+2];  /* UART送信文字列格納 */
    
    CyGlobalIntEnable; /* Enable global interrupts. */
    
    ADC_SAR_Seq_1_Start();
    ADC_SAR_Seq_1_IRQ_Enable();
    UART_1_Start();
    Pin_LED_Write(0);   /* BSoC内蔵の緑色 LED 消灯 = AD変換停止。 */
    state = STATE_STOP;
    /* PWM は始動しない。PWM を始動すると、AD変換を開始してしまう。 */
    
    for(;;)
    {
        /* 1回のAD変換完了に伴う、データ処理
         *
         * AD変換の完了をフラグ(adc_ready_flag>=1) で確認して、多チャンネルAD変換結果を PCへ UART送信する。
         *
         * AD変換は、自動生成API である ADC_SAR_Seq_1_INT.c 内の CY_ISR( ADC_SAR_Seq_1_ISR )
         * 割り込み関数内で実装する。AD変換タイミングは、ハードウェアトリガであり、TopDesign.cysch の AD変換
         * コンポーネントの SOC 端子立ち上がりエッジで割り込みが起動する。 CY_ISR( ADC_SAR_Seq_1_ISR )
         * 割り込み関数が呼び出される。
         *
         * CY_ISR( ADC_SAR_Seq_1_ISR ) 割り込み関数の処理 ----------------------------------
         *      AD変換結果を、グローバル変数 adc_result配列に格納する。最後にグローバル変数フラグ
         *      adc_done_flag を1加算する。
         *
         * main() 関数内での処理 -----------------------------------------------------------
         *      AD変換完了の知らせであるグローバル変数フラグ adc_done_flag が非0 であることを確認し
         *      て、多チャンネルのAD変換結果をCSV文字列形式で UART 送信する。最後にフラグ adc_done_flag
         *      を 0 に戻す。
         *
         * (注意) adc_ready_flag > 1 の時は、main() 関数内のデータ処理が遅く(=AD変換速度が速く)
         *        データサンプリング欠損を示唆している。UART通信速度が十分か、AD変換速度が速すぎないか
         *        要確認のこと。
         */
        if(adc_done_flag >= 1){
            /* AD変換結果 adc_result 配列の数値を、カンマ区切りの文字列に変換し、send_message に格納。 */
            str_index = 0;
            for(i=0; i<ADC_CHANNEL; ++i){
                str_index += sprintf(send_message+str_index, "%d,", adc_result[i]);
            }
            send_message[str_index-1] = '\n';     /* 行末のカンマを、ラインフィード(LF)に置換 */
            send_message[str_index] = '\0';       /* 文字列終端ヌル文字を付加 */
            UART_1_UartPutString(send_message);   /* ヌル文字終端の文字列を、UART送信 */
            
            /* ここまでの処理を経た後に、adc_done_flag > 1 の場合は、処理速度が不足していることを示唆する。 */
            if(adc_done_flag > 1){
                /* AD変換速度が速すぎて、データサンプリング欠損している場合の処理(本稿では未実装) */
                /* Do nothing */
            }
            adc_done_flag = 0;                    /* 最後にAD変換完了フラグを 0 に戻す */
        }

        /* PCからの命令受信処理 何も受信していなければ command = 0 */
        command = (uint8)(UART_1_UartGetChar() & 0x000000FFl);
        switch(command){
            /* AD変換開始命令(ASCIIコード 'a')を受領した場合 */
            case COMMAND_START:
                if(state == STATE_STOP){
                    state = STATE_START;  /* 現在のステートを STATE_START(AD変換開始) にする。 */
                    PWM_1_Start();        /* AD変換開始は、PWM の開始と等価(PWMからの割り込みでAD変換) */
                    Pin_LED_Write(1);     /* BSoC内蔵の緑色 LED 点灯 */
                }
                break;
            /* AD変換停止命令(ASCIIコード 'z')を受領した場合 */
            case COMMAND_STOP:
                if(state == STATE_START){
                    state = STATE_STOP;   /* 現在のステートを STATE_STOP(AD変換停止) にする。 */
                    PWM_1_Stop();         /* AD変換停止は、PWM の停止と等価。 */
                    Pin_LED_Write(0);     /* BSoC内蔵の緑色 LED 消灯。 */
                }
                break;
            default:
                /* Do nothing */
                break;
        }
    }
}
