﻿# -*- coding: utf-8 -*-
"""汎用オシロスコープアプリケーション by Python

本誌記事「PSoC x Python お手軽オシロスコープの製作」
の汎用オシロスコープアプリケーションです。2ch のアナログ入力
を持ち、1000Hz のサンプリング周波数※とします。
GUI画面のボタンを押すことで、PSoC への AD変換開始/停止
命令を発行し、オシロスコープの掃引オン/オフを切り替えます。

※厳密には、本アプリケーションはサンプリング周波数に依存しません。
 一回のサンプル 時間のデータは、改行(LF) で識別できるためです。
 単にグラフ の横軸を定義するための参考情報です。


【入力データフォーマット】
時系列多チャンネルデータが、CSVフォーマットで入力される前提とします。
一回のサンプル時間ごとのデータは、改行コード(LF) で区切ります。
多チャンネルのデータは、カンマ区切りの CSV フォーマットで表現します。
一つ一つのデータは、10進数/小数(整数も可能)/正負符号可 とします。

(例) 3ch の時系列データ
10,39.6,-10.8\n11.2,38,-9\n12,37.2,-9.5\n...
---
1回目のサンプリング 1ch:10, 2ch:39.6, 3ch:-10.8
2回目のサンプリング 1ch:11.2, 2ch:38, 3ch:-9
3回目のサンプリング 1ch:12, 2ch:37.2, 3ch:-9.5


【一般的な使い方】
Windows, Linux(含Raspbian OS), Mac OS の場合は、PSoC
との接続に USB-UART変換モジュールを使用します。秋月電子通商の
「超小型USBシリアル変換モジュール」 が便利です。
(http://akizukidenshi.com/catalog/g/gM-08461/)
変換モジュールの TX端子を、PSoC の UART-RX端子に接続します。
変換モジュールの RX端子を、PSoC の UART-TX端子に接続します。
変換モジュールの +5V端子から、PSoC の電源を供給します。

USBシリアル変換モジュールを使う場合は、ソースコードの 「設定定数」欄の
DEVICE_NAME = None とします。デバイスを自動で検索してオープンします。


【オンボードPC(Raspberry Pi) 固有の使い方】
Raspberry Pi(Raspbian OS) の UART 端子を使って、PSoC と
直接UART 通信することも可能です。
① Raspberry Pi の GPIO14 端子を、PSoC の UART-RX 端子と接続します。
② Raspberry Pi の GPIO15 端子を、PSoC の UART-TX 端子と接続します。
③ Raspberry Pi の GPIO14, 15 は、初期状態では他の目的に使用
  されています。UART 機能として使う場合には、設定変更が必要です。
  OSバージョンや、機種毎に設定が異なり、多少複雑なので、ネット等で検索して設定してください。
④ ソースコードの 「設定定数」欄の DEVICE_NAME を、UART デバイス名に書換えます。
  DEVICE_NAME = 'dev/ttyAMA0'


[Windows10 確認環境]
Python      3.7.0
Numpy       1.15.1
Matplotlib  2.2.3
Pyseiral    3.4

[Raspbian OS 確認環境]
(Raspberry Pi 3)
Python      3.5.3
Numpy       1.15.4
Matplotlib  3.0.2
Pyseiral    3.2.1

[Ubuntu 16.04 確認環境]
Python      3.5.2
Numpy       1.13.3
Matplotlib  3.0.0
Pyseiral    3.3

※Raspberry Pi Zero W では、AD変換のサンプリング周波数 1000Hz の高速オシロスコープ表示
 が出来ません。CPU の処理能力が不足しています。AD変換のサンプリング周波数 50Hz 以下が実用
 限界になります。

本ソースコードはフリー・ソフトウェアであり、また無保証です。
用途に制限はありません。あなたの責任の下において、個人的・非営利的な
ものから商用製品の開発に及ぶ目的に使用・改変・再配布することができます。
"""
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.widgets import Button
import serial
from ringbuffer import RingBuffer
from serial.tools import list_ports

# 設定定数 (自由に変更可能です。) *********************************
WINDOW_X = 9         # ウィンドウサイズ横 [inch]
WINDOW_Y = 6         # ウィンドウサイズ縦 [inch]
LABEL_X = 'Time[sec]'   # x軸ラベル
LABEL_Y = 'ADC Value'   # y軸ラベル
CH_SIZE = 2          # データチャンネル数 = グラフプロット本数
# プロット線の色 1ch:茶, 2ch:赤, 3ch:橙, 4ch:黄,
#           5ch:緑, 6ch:青, 7ch:紫, 8ch:灰
CH_COLOR = ('#993300', '#ff0000', '#ff6600', '#ffff00',
            '#00ff00', '#0000ff', '#800080', '#969696')
SAMPLE_RATE = 1000   # サンプリング周波数[Hz] (横軸の目盛り作成目的)
PLOT_POINTS = 3000   # 横軸プロット点数
RANGE_Y_MAX = 2000   # 縦軸スケール最大値
RANGE_Y_MIN = -2000  # 縦軸スケール最小値
# UART 通信条件は、マイコンの UART コンポーネントの設定条件に合わせること
SERIAL_OPTIONS = {'baudrate': 115200,
                  'bytesize': serial.EIGHTBITS,
                  'parity': serial.PARITY_NONE,
                  'stopbits': serial.STOPBITS_ONE}
# 通信デバイス名(UART 又は USBシリアル変換モジュールのデバイス名) が分かっている場合は、
# DEVICE_NAME に記述する。
# (ex. Windows の USBシリアル変換モジュールの場合 "COM48" 等...COM名
#      Linux の場合 "/dev/ttyAMA0" 等...デバイスファイル名)
# DEVICE_NAME を None にすると、USBシリアル変換モジュール(FTDI社製IC搭載)
# を使っているとみなし、USBデバイスを自動検索する。(Windows, Linux, Mac 共通)
DEVICE_NAME = None
USB_VID = 0x0403     # USBシリアル変換モジュールで使われるIC の USB Vendor ID
# プロットと同時にデータファイル保存したい場合、ファイル名を記述。 None にすると、ファイル保存しない。
FILE_NAME = r'log.csv'

# グローバル変数 (変更不要) ***************************************
plot_line = [None] * CH_SIZE    # Line2D オブジェクト格納予定
serial_object = None   # シリアル通信オブジェクト。 None はデバイスクローズ状態。
receive_str = ''       # 受信文字列バッファ
fp = None              # セーブファイルオブジェクト
# データ保存用リングバッファ 余裕を見て x軸プロット点数の2倍点 x chサイズ分 のリングバッファ
ring_buffer = RingBuffer(PLOT_POINTS*2, CH_SIZE)


def usb_devicename(vid=0x0403, pid=None):
    """VID, PID から、USBデバイス名を捜索

    引数で指定した Vendor ID, Product ID に合致する
    USB機器を捜索して、あればデバイス名を返す。無ければ、
    None が返る。
    例)  Windows: "COM24", Linux: "/dev/ttyAMA0"

    引数
    ----------
    vid : 捜索したUSB機器の VID(Vendor ID) 番号。※1
    pid : 捜索したUSB機器の PID(Product ID) 番号

    両方(VID,PID) 指定も出来るが、一方を None値としても良い。
    Noneとした方は不問とする。

    戻値
    ----------
    該当するUSB機器のデバイス名

    ※1
    世の中で主流の「USBシリアル変換モジュール」には、FTDI社のICチップが
    使われています。同社の Vendor ID は 0x0403 です。デフォルト
    引数とします。
    """
    # dev_list には、ListPortInfoオブジェクトのリストが返る。
    # ListPortInfo は、接続されている シリアルデバイスの情報を持っている。
    # 接続されているシリアルデバイスの数だけ、リストで返る。
    dev_list = list_ports.comports()
    dev_name = None
    if((vid is not None) and (pid is None)):
        for i in dev_list:
            if(i.vid == vid):
                dev_name = i.device
                break
    elif((vid is None) and (pid is not None)):
        for i in dev_list:
            if(i.pid == pid):
                dev_name = i.device
                break
    elif((vid is not None) and (pid is not None)):
        # dev_list はリストなので、要素を順に for で抽出。
        # i は ListPortInfoオブジェクトになる。
        for i in dev_list:
            # ListPortInfo　オブジェクトは、以下のメンバ変数を有している。
            #    pid (int型) : USB機器である場合、PID(Product ID)番号。
            #    vid (int型)　：　USB機器である場合、VID(Vendor ID)番号。
            #    device (str型) : デバイス名
            if(i.vid == vid and i.pid == pid):
                dev_name = i.device
                break
    return(dev_name)


def callback_btn(event):
    """ボタン押下イベント関数

    シリアル通信デバイス名が既知の場合は冒頭で DEVICE_NAME を非None で定義。
    もし None であれば、USB-UART変換冶具と見なし、FTDI社の USBデバイス
    を捜索する。
    シリアル通信デバイスがオープンされると、定期呼出し関数 anime_enent() の影響で、
    グラフプロットが再描画される。但しデバイスが見つからなければ、オープンしない。

    シリアル通信デバイスがオープン済みなら、クローズする。クローズされると、グラフプロット
    は定期呼出し関数 anime_enent() 内で再描画されず停止する。
    """
    global serial_object, receive_str, ring_buffer, btn, fp
    global DEVICE_NAME, USB_VID, SERIAL_OPTIONS, FILE_NAME

    # シリアル通信デバイスがオーブンされていなければ、デバイス捜索 & デバイスオープン。
    # ボタン名・色を変え、マイコンにAD変換開始コマンドを送る。各種保存用データ
    # をクリアし、プロット準備をする。
    if(serial_object is None):
        if(DEVICE_NAME is None):
            # VID=0x0403(FTDI社のVID) を捜索して、デバイス名取得
            search_device = usb_devicename(USB_VID)
        else:
            search_device = DEVICE_NAME
        # シリアル通信デバイスが見つかれば、USBオープン。
        if(search_device is not None):
            try:
                serial_object = serial.Serial(
                        search_device, timeout=0, **SERIAL_OPTIONS)
            except:
                serial_object = None
            else:
                btn.color = 'skyblue'
                btn.label.set_text('Plotting')
                serial_object.flushInput()   # シリアル通信バッファクリア
                serial_object.write(b'a')    # PSoC へAD変換開始命令
                receive_str = ''             # マイコンからの受信文字列を蓄積保存
                ring_buffer.clear()          # データ保存用リングバッファのクリア
                if(FILE_NAME is not None):   # データのファイル保存をする場合、ファイルオープン
                    try:
                        fp = open(FILE_NAME, mode='w')
                    except:
                        fp = None
    # 既に USBがオープン中なら、USBクローズする。
    else:
        serial_object.write(b'z')    # PSoC へAD変換停止命令
        serial_object.flushInput()
        receive_str = ''
        serial_object.close()
        serial_object = None
        btn.color = 'mistyrose'
        btn.label.set_text('Waiting')
        if(fp is not None):          # データのファイル保存している場合、ファイルクローズ処理
            fp.close()
            fp = None


def anime_event(frame_number):
    """animation.FuncAnimation() から定期的に呼ばれる関数

    通信デバイスがオープンされていたら、CSV形式の受信データを受取り、
    全チャンネルデータを数値リスト化して、保存用リングバッファに蓄積。
    同バッファから最新の数値データを取りだし、Line2D オブジェクトデータを
    更新する。すると、自動でグラフプロットが再描画される。
    """
    global plot_line, serial_object, receive_str, ring_buffer, fp
    global PLOT_POINTS, CH_SIZE

    # シリアル通信デバイスがオープン(非None) の場合には、データ受信しグラフ更新
    if(serial_object is not None):
        # グローバル変数の文字列バッファに USB受信文字列を追記
        # 受信データは改行仕切りされた、CSV形式。
        # 例) "10,39.6,-10.8\n11.2,38,-9\n..."
        try:
            receive_str += serial_object.read_all().decode()
        except:
            # 通信エラー発生時は、強制的に停止処理を呼び出す。
            callback_btn(None)
        # 改行文字('\n') を捜索して、その配列位置を取得。無ければ -1 が返る。
        find_index = receive_str.find('\n')
        while(find_index != -1):
            # データのファイル保存が有効な場合(改行文字も含めて保存)
            if(fp):
                fp.write(receive_str[: find_index+1])

            # 改行文字以前を取りだし、カンマで分離。分離された文字列リストを取得
            # 例) "10,39.6,-10.8" -> ["10", "39.6", "-10.8"]
            split_str = receive_str[: find_index].split(',')
            # 文字列バッファを、捜索された改行文字以降に差し替え
            receive_str = receive_str[find_index+1:]
            # 上記分離された文字列リストを、数値リストに変換
            # 例) ["10", "39.6", "-10.8"] --> [10, 39.6, -10.8]
            convert_float = [float(i) for i in split_str]
            # 得られた数値リストを、保存用リングバッファに追記
            ring_buffer.push(convert_float)
            # さらに改行文字があれば、再検索し繰り返す。
            find_index = receive_str.find('\n')

        # 保存用リングバッファから最新データ取得し、AD変換軸(y軸) Line2D データ更新。
        # 時間軸(x軸) は不変なので、更新しない。
        latest_data = ring_buffer.pop_latest(PLOT_POINTS)
        for i in range(CH_SIZE):
            # i-ch目の PLOT_POINTS点データは、配列 [:, i] に格納されている。
            # Line2D オブジェクトの y軸データを更新する set_ydata()
            plot_line[i].set_ydata(latest_data[:, i])

    # シリアル通信デバイスが停止中は、グラフを更新しない。
    else:
        pass

    # animation.FuncAnimation の blit引数を True にして高速描画の場合、
    # 更新される全ての Artistsオブジェクト(ここではプロット線：Line2Dオブジェクト) の
    # リストを return する必要がある。 blit=False では return は無視される。
    # よって、以下のリターン文を記述しておく。
    return(plot_line)


# 大元の Figureオブジェクト生成
fig = plt.figure(figsize=(WINDOW_X, WINDOW_Y))
fig.canvas.set_window_title('Log View')    # ウィンドウタイトルバーの表示文字
ax_plot = fig.subplots(nrows=1, ncols=1)   # fig に属する、1行1列の Axesオブジェクト生成、登録
ax_plot.set_xlabel(LABEL_X)       # x軸ラベル設定
ax_plot.set_ylabel(LABEL_Y)       # y軸ラベル設定
# subplotエリアのサイズ微調整。フルスケールを 0～1 とした数値指定。
fig.subplots_adjust(top=0.95, bottom=0.15, left=0.15, right=0.95)
# 軸の範囲設定
ax_plot.set_xlim(0, PLOT_POINTS/SAMPLE_RATE)
ax_plot.set_ylim(RANGE_Y_MIN, RANGE_Y_MAX)
# Figureオブジェクトに属する、新しい Axesオブジェクトを生成
# Axesオブジェクトの配置位置とサイズ [left, bottom, width, height]
# 左下を (0, 0) として、ウィンドウフルスケールを 1 とした数値指定。
ax_btn = fig.add_axes([0.05, 0.02, 0.15, 0.05])
# 直前に作成した Axesオブジェクトに属する、Button オブジェクトを生成
btn = Button(ax_btn, label='Waiting', color='mistyrose')
# Buttonオブジェクトに、押された場合のコールバック関数を登録
btn.on_clicked(callback_btn)

# 時間軸(x軸)データ
time_data = [i/SAMPLE_RATE for i in range(PLOT_POINTS)]
# AD変換軸(y軸)データ。最初は 0 になる。
first_data = ring_buffer.pop_latest(PLOT_POINTS)
for i in range(CH_SIZE):
    # plot関数 はプロット線を描き、Line2Dオブジェクト = プロット線を示すオブジェクトのリストを返す。
    # ここでは1本のプロット線毎に plot関数を呼び出すので、plot関数の返値は要素1 のリスト。
    # plot_line[i] に iチャンネル目の Line2D オブジェクトを代入する。
    plot_line[i] = (ax_plot.plot(time_data, first_data[:, i],
                                 color=CH_COLOR[i],
                                 linestyle='solid'))[0]    # 要素1のリストを取り出す。
# 定期的に呼び出される関数を登録し、同関数内でアニメーション効果を生み出す
#    第一引数 : figureオブジェクト , 第二引数 : 定期的に呼び出される関数登録
#    interval引数 : 呼出し間隔 [msec]
#    blit引数 : Trueで高速描画。但し本用途では、特に不要なので、False 指定。
ani = animation.FuncAnimation(fig, anime_event, interval=20, blit=True)
plt.show()
# ウィンドウが閉じられるまで、plt.show() に留まり続ける。

# 終了処理。シリアル通信デバイスがオープンされていたら、PSoCにAD変換停止命令し、デバイスクローズ。
# ファイルがオープンされていたら、ファイルクローズ。
if(serial_object is not None):
    serial_object.write(b'z')    # PSoCにAD変換停止命令
    serial_object.close()
if(fp is not None):
    fp.close()
